/* aloop.c - performance checker */

/* Bucari, Prosum, 2009 */

#if HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <atm.h>
#include <pthread.h>
#include <time.h>

#define LOCK(m) pthread_mutex_lock(&(m));
#define UNLOCK(m) pthread_mutex_unlock(&(m));

struct sockaddr_atmpvc addr;
struct atm_qos qos;
unsigned long sendcpt, recvcpt;
unsigned long *buf;
unsigned pcr, us;
int type = 1, gfc = 0, clp = 0;
int sduSize;

pthread_mutex_t mx = PTHREAD_MUTEX_INITIALIZER;
int s;

static void usage(const char *name)
{
    fprintf(stderr,"usage: %s [itf.]vpi.vci [qos spec]\n", name);
    exit(1);
}

void * recvTask (void * nothing)
{
    int size;

    while (1) {
        size = read(s, buf, sduSize);
        if (size < 0) {
            perror("read");
            exit (1);
        }
        if (size != sduSize)         
          fprintf(stderr,"bad RX (%d)\n", size);
        
        LOCK(mx);
        recvcpt++;
        UNLOCK(mx);
    }
    return NULL;
}

void * sendTask (void * nothing)
{
    while (1) {
        if (qos.aal == ATM_AAL0) {
            buf[0] = (gfc << ATM_HDR_GFC_SHIFT) | (addr.sap_addr.vpi << ATM_HDR_VPI_SHIFT) |
            (addr.sap_addr.vci << ATM_HDR_VCI_SHIFT) | (type << ATM_HDR_PTI_SHIFT) | clp;
        }
        
        if (write(s, buf, sduSize) != sduSize) {    /* the length does not matter below 52 since */
                                          /* the smallest transmitted packet is one cell */
            perror("write");
        }
        LOCK(mx);
        sendcpt++;
        UNLOCK(mx);
    }
    return NULL;
}

int main(int argc,char **argv)
{
    pthread_t recvThread = 0, sendThread = 0;
    pthread_attr_t attr;
    struct timespec to = {1, 0};
    unsigned long sw, rw;
    int ncell;
    
    sendcpt = recvcpt =0;

    if (argc < 2 || argc == 3) 
        usage(argv[0]);

    if ((s = socket(PF_ATMPVC, SOCK_DGRAM,0)) < 0) {
        perror("socket");
        exit (1);
    }

    memset(&addr,0,sizeof(addr));

    if (text2atm(argv[1], (struct sockaddr *) &addr,sizeof(addr), T2A_PVC | T2A_UNSPEC | T2A_WILDCARD) < 0) 
        usage(argv[0]);

    memset(&qos,0,sizeof(qos));

    /* set default values */
    qos.aal = ATM_AAL0;
    qos.txtp.traffic_class = ATM_CBR;
    qos.txtp.pcr = 90000;
    qos.txtp.max_sdu = 52;                  /* must be 52 for aal0 */
    qos.rxtp = qos.txtp;

    if (argc > 3 && !strcmp(argv[2], "qos")) {
        if (text2qos(argv[3], &qos, 0)) {
            printf ("QoS specification error\n");
            exit (1);
        } 
    }

    if (qos.aal == ATM_AAL0) {
        sduSize = 52;
        qos.rxtp.max_sdu = qos.txtp.max_sdu = 52;
        ncell = 1;
    }
    else if (qos.aal == ATM_AAL5) {
        if (qos.txtp.max_sdu < 48)
            qos.txtp.max_sdu = 48;

        qos.rxtp.max_sdu = qos.txtp.max_sdu;
        sduSize = qos.txtp.max_sdu - ATM_AAL5_TRAILER;
        ncell = (sduSize + 51)/52; 
    }
    else {
        printf ("Unsupported AAL\n");
        exit (1);
    }
    
    buf = (ulong *)malloc (qos.txtp.max_sdu);     
            
    if (setsockopt(s, SOL_ATM, SO_ATMQOS, &qos, sizeof(qos)) < 0) {
        perror("setsockopt SO_ATMQOS");
        exit(1);
    }
    if (bind(s,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
        perror("bind");
        exit(1);
    }

    printf ("Running AAL%c, %s, pcr = %d, packet size = %d (%d cells)\n",
       qos.aal == ATM_AAL5? '5': '0',  
       qos.txtp.traffic_class == ATM_CBR? "CBR": 
       qos.txtp.traffic_class == ATM_UBR? "UBR": 
       qos.txtp.traffic_class == ATM_VBR? "VBR": "not supported",
       qos.txtp.pcr, qos.txtp.max_sdu, ncell); 
            
    pthread_attr_init (&attr);
    pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);

    /* create sendTask */
    pthread_create (&recvThread, &attr, recvTask, NULL);

    /* create recvTask */
    pthread_create (&sendThread, &attr, sendTask, NULL);
    
    while (1) {
        clock_nanosleep(CLOCK_REALTIME, 0, &to, NULL);
        LOCK (mx);
        sw = sendcpt * ncell;
        rw = recvcpt * ncell;
        sendcpt = recvcpt = 0;
        UNLOCK(mx);
        fprintf (stderr, "Cell Rate: TX = %lu, RX = %lu\n", sw, rw);
    }    
}

